@@ -1,20 +1,20 @@
 # refinement.py - routines to improve geometry.
-import mdplus
 from mdplus import fast
 import numpy as np
 from scipy.optimize import minimize
 from mdplus import utils
 
+
 class ENM(object):
 
     def __init__(self, max_its=1000, tol=0.001, fc=0.1):
         """
-        A Elastic Network-based Coordinate Optimiser. 
+        A Elastic Network-based Coordinate Optimiser.
 
         Adjusts the coordinates in each snapshot in a trajectory to
         optimise a set of distance restraints. The restraints may be
         provided explicitly or identified automatically from analysis
-        of short distances in the structures in a training trajectory. 
+        of short distances in the structures in a training trajectory.
 
         The automatic method assumes the trajectory is of a single molecule
         and so, in graph-speak, is a single connected component.
@@ -52,7 +52,7 @@ class ENM(object):
             restraints: [N_bonds, 3] array of "bonded" atoms. Each row is
                         [i, j, r] where i and j are indices (cast to floats)
                         and r is the target distance.
-                
+
         """
         X = utils.check_dimensions(X, ensure_traj=True)
         self.n_atoms = X.shape[1]
@@ -64,12 +64,38 @@ class ENM(object):
             ik = [[a[0], a[2]] for a in angles]
             restraints.extend(ik)
             self.restraints = np.array(restraints, dtype=np.int32)
-            k = utils.pair_ind_to_dist_ind(self.n_atoms, self.restraints[:, 0], self.restraints[:, 1])  
+            k = utils.pair_ind_to_dist_ind(self.n_atoms, self.restraints[:, 0],
+                                           self.restraints[:, 1])
             self.d_ref = dm[k]
         else:
             self.restraints = np.array(restraints[:, :2], dtype=np.int32)
             self.d_ref = np.array(restraints[:, 2], dtype=np.float32)
 
+    def get_state(self):
+        """
+        Get the state of the transformer as a dictionary
+
+        """
+        if self.n_atoms is None:
+            raise RuntimeError("Error - the transformer has not been trained")
+        config = {}
+        for key in ['max_its', 'tol', 'fc', 'n_atoms']:
+            config[key] = self.__dict__[key]
+        config['restraints'] = self.restraints.tolist()
+        config['d_ref'] = self.d_ref.tolist()
+        config['result'] = None
+        return config
+
+    def set_state(self, config):
+        """
+        Set the state of the transformer from a dictionary
+
+        """
+        for key in ['max_its', 'tol', 'fc', 'n_atoms', 'result']:
+            self.__dict__[key] = config[key]
+        self.restraints = np.array(config['restraints'], dtype=np.int32)
+        self.d_ref = np.array(config['d_ref'], dtype=np.float32)
+
     def transform(self, X):
         """
         Refine sets of coordinates.
@@ -85,16 +111,26 @@ class ENM(object):
         if one_frame:
             X = X.reshape((1, -1, 3))
         if X.shape[1] != self.n_atoms:
-            raise ValueError('Error: number of atoms in array to refine ({}) does not match number used to train refiner ({})'.format(X.shape[1], self.n_atoms))
+            raise ValueError(f'Error: number of atoms in array'
+                             f'to refine ({X.shape[1]})'
+                             f' does not match number used to'
+                             f' train refiner ({self.n_atoms})')
         Xout = []
         for Xc in X:
-            self.result = minimize(fast.fg, Xc.flatten().astype(np.float32), method='TNC', jac=True, args=(self.restraints, self.d_ref.astype(np.float32), self.fc), options={'maxfun': self.max_its, 'ftol': self.tol})
-            Xout.append(self.result.x.reshape((-1, 3))) 
+            self.result = minimize(fast.fg, Xc.flatten().astype(np.float32),
+                                   method='TNC', jac=True,
+                                   args=(self.restraints,
+                                         self.d_ref.astype(np.float32),
+                                         self.fc),
+                                   options={'maxfun': self.max_its,
+                                            'ftol': self.tol})
+            Xout.append(self.result.x.reshape((-1, 3)))
         Xout = np.array(Xout)
         if one_frame:
             Xout = Xout[0]
         return Xout
 
+
 def fix_bumps(X, dmin, dmax):
     """
     Fix bumps in coordinate array x using ENM approach.
@@ -115,7 +151,7 @@ def fix_bumps(X, dmin, dmax):
     if one_frame:
         X = X.reshape((1, -1, 3))
     Xout = np.zeros_like(X)
-    for l, Xc in enumerate(X):
+    for iframe, Xc in enumerate(X):
         ij = utils.get_bumps(Xc, dmax)
         n_bumps = len(ij)
         rmin = np.zeros(n_bumps)
@@ -129,7 +165,7 @@ def fix_bumps(X, dmin, dmax):
         restraints[:, 2] = rmin
         minimiser = ENM()
         minimiser.fit(Xc, restraints=restraints)
-        Xout[l] = minimiser.transform(Xc)
+        Xout[iframe] = minimiser.transform(Xc)
     if one_frame:
         Xout = Xout[0]
     return Xout
